import bpy
import mathutils
import itertools

class OBJECT_OT_create_bounce_curve(bpy.types.Operator):
    """Create bounce curve between selected objects"""
    bl_idname = "object.create_bounce_curve"
    bl_label = "Create Bounce Curve"
    bl_options = {'REGISTER', 'UNDO'}

    height: bpy.props.FloatProperty(
        name="Curve Height",
        description="Height adjustment for the curve",
        default=0.005,
        min=0.0,
        max=1.0,
        subtype='DISTANCE',  # Displays units in the slider
    )

    def execute(self, context):
        selected_objects = context.selected_objects
        positions = [obj.location.copy() for obj in selected_objects]
        if len(positions) < 2:
            self.report({'WARNING'}, "Select at least two objects.")
            return {'CANCELLED'}

        order = compute_shortest_path(positions)
        ordered_positions = [positions[i] for i in order]
        curve_obj = create_bounce_curve(ordered_positions, self.height)
        self.report({'INFO'}, "Bounce curve created.")
        return {'FINISHED'}

def compute_shortest_path(positions):
    n = len(positions)
    if n <= 10:
        # For small n, compute the exact shortest path using permutations
        return compute_shortest_path_brute_force(positions)
    else:
        # For larger n, use an improved nearest-neighbor heuristic
        return compute_nearest_neighbor_path(positions)

def compute_shortest_path_brute_force(positions):
    n = len(positions)
    indices = range(n)
    min_total_distance = None
    best_path = None
    for perm in itertools.permutations(indices):
        total_distance = 0
        for i in range(n - 1):
            dist = (positions[perm[i + 1]] - positions[perm[i]]).length
            total_distance += dist
        if min_total_distance is None or total_distance < min_total_distance:
            min_total_distance = total_distance
            best_path = perm
    return best_path

def compute_nearest_neighbor_path(positions):
    n = len(positions)
    min_total_distance = None
    best_path = None
    for start in range(n):
        indices = set(range(n))
        path = []
        current = start
        indices.remove(current)
        path.append(current)
        total_distance = 0
        while indices:
            current_pos = positions[current]
            min_dist = None
            next_index = None
            for i in indices:
                dist = (positions[i] - current_pos).length
                if min_dist is None or dist < min_dist:
                    min_dist = dist
                    next_index = i
            if next_index is None:
                break  # Safety check in case of an error
            total_distance += min_dist
            indices.remove(next_index)
            path.append(next_index)
            current = next_index
        # Compare total distance
        if min_total_distance is None or total_distance < min_total_distance:
            min_total_distance = total_distance
            best_path = path
    return best_path

def create_bounce_curve(ordered_positions, height):
    # Prepare points list with midpoints added and adjusted
    points = []
    handle_types = []

    for i in range(len(ordered_positions) - 1):
        # Original point
        p0 = ordered_positions[i]
        points.append(p0)
        handle_types.append('VECTOR')  # Original points have 'VECTOR' handles

        # Midpoint
        p1 = ordered_positions[i + 1]
        mid_point = (p0 + p1) / 2
        mid_point.z += height  # Move midpoint up by 'height' units on Z-axis
        points.append(mid_point)
        handle_types.append('AUTO')  # Midpoints have 'AUTO' handles

    # Add the last original point
    points.append(ordered_positions[-1])
    handle_types.append('VECTOR')

    # Reverse the points and handle_types to reverse the curve direction
    points.reverse()
    handle_types.reverse()

    # Create a new curve
    curve_data = bpy.data.curves.new(name='SimpleBounceCurve', type='CURVE')
    curve_data.dimensions = '3D'
    spline = curve_data.splines.new('BEZIER')
    spline.bezier_points.add(len(points) - 1)

    # Assign points and handle types
    for i, (point, handle_type) in enumerate(zip(points, handle_types)):
        bp = spline.bezier_points[i]
        bp.co = point
        bp.handle_left_type = handle_type
        bp.handle_right_type = handle_type

    # Create a new object with the curve
    curve_obj = bpy.data.objects.new('SimpleBounceCurveObject', curve_data)
    # Add the object to the scene
    bpy.context.collection.objects.link(curve_obj)
    return curve_obj

def menu_func(self, context):
    self.layout.operator(OBJECT_OT_create_bounce_curve.bl_idname)

def register():
    bpy.utils.register_class(OBJECT_OT_create_bounce_curve)
    bpy.types.VIEW3D_MT_object.append(menu_func)

def unregister():
    bpy.utils.unregister_class(OBJECT_OT_create_bounce_curve)
    bpy.types.VIEW3D_MT_object.remove(menu_func)

if __name__ == "__main__":
    register()
